from tkinter import RIGHT, BOTH, Tk, Frame, Button

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from diffeq_solver_tk import tkinter_config as config
from diffeq_solver_tk.finite_difference import solve_first_order_ode, solve_second_order_ode, solve_heat_equation, \
    solve_wave_equation
from diffeq_solver_tk.forms.differential_equation_form import DifferentialEquationForm
from diffeq_solver_tk.forms.first_order_ode_form import FirstOrderODEForm
from diffeq_solver_tk.forms.heat_equation_form import HeatEquationForm
from diffeq_solver_tk.forms.second_order_ode_form import SecondOrderODEForm
from diffeq_solver_tk.forms.wave_equation_form import WaveEquationForm
from diffeq_solver_tk.differential_equation_service import OrdinaryDifferentialEquationService, \
    BoundedEquationService


class MainView(Frame):
    first_order_ode_form: FirstOrderODEForm
    second_order_ode_form: SecondOrderODEForm
    heat_equation_form: HeatEquationForm
    wave_equation_form: WaveEquationForm
    selected_form: DifferentialEquationForm = None

    def __init__(self, app: Tk):
        Frame.__init__(self, master=app)
        self.initialize_forms()
        self.build_nav_bar()
        self.build_details_container()
        self.handle_select_form(self.first_order_ode_form)

    def initialize_forms(self):
        figure = Figure(figsize=(6, 1), dpi=91.4)
        canvas = FigureCanvasTkAgg(figure, self)
        canvas.get_tk_widget().pack(side=RIGHT, fill=BOTH)
        self.first_order_ode_form = FirstOrderODEForm(
            self, canvas,
            OrdinaryDifferentialEquationService(solve_first_order_ode, figure))
        self.second_order_ode_form = SecondOrderODEForm(
            self, canvas,
            OrdinaryDifferentialEquationService(solve_second_order_ode, figure))
        self.heat_equation_form = HeatEquationForm(
            self, canvas,
            BoundedEquationService(solve_heat_equation, figure))
        self.wave_equation_form = WaveEquationForm(
            self, canvas,
            BoundedEquationService(solve_wave_equation, figure))

    def build_nav_bar(self):
        nav_bar_frame = Frame(self)
        nav_bar_frame.pack(side="top", fill="x", expand=False)
        self.place_nav_bar_button(nav_bar_frame, "First Order ODE", self.first_order_ode_form, 0)
        self.place_nav_bar_button(nav_bar_frame, "Second Order ODE", self.second_order_ode_form, 1)
        self.place_nav_bar_button(nav_bar_frame, "Heat Equation", self.heat_equation_form, 2)
        self.place_nav_bar_button(nav_bar_frame, "Wave Equation", self.wave_equation_form, 3)

    def build_details_container(self):
        container = Frame(self)
        container.configure(background=config.details_background)
        container.pack(side="top", fill="both", expand=True)
        self.first_order_ode_form.place(in_=container, x=0, y=config.details_top_margin, relwidth=1, relheight=1)
        self.second_order_ode_form.place(in_=container, x=0, y=config.details_top_margin, relwidth=1, relheight=1)
        self.heat_equation_form.place(in_=container, x=0, y=config.details_top_margin, relwidth=1, relheight=1)
        self.wave_equation_form.place(in_=container, x=0, y=config.details_top_margin, relwidth=1, relheight=1)

    def place_nav_bar_button(self, frame: Frame, text: str, target_form: DifferentialEquationForm, column: int):
        Button(frame,
               text=text,
               font=config.nav_bar_font,
               foreground=config.nav_bar_foreground,
               background=config.nav_bar_background,
               command=lambda: self.handle_select_form(target_form),
               width=config.nav_bar_button_width).grid(row=0, column=column)

    def handle_select_form(self, form: DifferentialEquationForm):
        if self.selected_form == form:
            return
        if self.selected_form:
            self.selected_form.reset()
        self.selected_form = form
        self.selected_form.lift()



from enum import auto, Enum
from tkinter import Button, Entry, Variable, Frame
from typing import Union, Dict

import diffeq_solver_tk.messages.common_messages as common_messages
import diffeq_solver_tk.messages.wave_equation_messages as messages
from diffeq_solver_tk.differential_equation_metadata import BoundaryConditions, BoundaryCondition, WaveEquationMetadata, BoundaryType
from diffeq_solver_tk.differential_equation_service import BoundedEquationService
from diffeq_solver_tk.forms.differential_equation_form import DifferentialEquationForm
from diffeq_solver_tk.forms.equation_form_builder import EquationFormBuilder


class WaveEquationFields(Enum):
    LEFT_BOUNDARY_TYPE = auto()
    RIGHT_BOUNDARY_TYPE = auto()
    LEFT_BOUNDARY_VALUES = auto()
    RIGHT_BOUNDARY_VALUES = auto()
    INITIAL_VALUES = auto()
    INITIAL_DERIVATIVES = auto()
    SOURCE = auto()
    WAVE_SPEED = auto()
    ALPHA = auto()
    LENGTH = auto()
    TIME = auto()
    SAMPLES = auto()


class WaveEquationForm(DifferentialEquationForm):
    field_entry_map: Dict[WaveEquationFields, Union[Entry, Variable]]
    equation_service: BoundedEquationService
    solve_button: Button
    export_button: Button
    render_plot_button: Button
    toggle_animation_button: Button

    def __init__(self, frame: Frame, canvas, equation_service: BoundedEquationService):
        DifferentialEquationForm.__init__(self, frame, canvas, equation_service)

    def build_form(self):
        builder = EquationFormBuilder[WaveEquationFields](self)
        self.build_entries(builder)
        self.build_buttons(builder)

    def build_entries(self, builder: EquationFormBuilder[WaveEquationFields]):
        builder.build_boundary_type_section(WaveEquationFields.LEFT_BOUNDARY_TYPE,
                                            WaveEquationFields.RIGHT_BOUNDARY_TYPE)
        builder.build_entry_row(WaveEquationFields.LEFT_BOUNDARY_VALUES,
                                messages.left_boundary_values,
                                messages.left_boundary_values_symbol, 2)
        builder.build_entry_row(WaveEquationFields.RIGHT_BOUNDARY_VALUES,
                                messages.right_boundary_values,
                                messages.right_boundary_values_symbol, 3)
        builder.build_entry_row(WaveEquationFields.INITIAL_VALUES,
                                messages.initial_values,
                                messages.initial_values_symbol, 4)
        builder.build_entry_row(WaveEquationFields.INITIAL_DERIVATIVES,
                                messages.initial_derivatives,
                                messages.initial_derivatives_symbol, 5)
        builder.build_entry_row(WaveEquationFields.SOURCE,
                                messages.source_term,
                                messages.source_term_symbol, 6)
        builder.build_entry_row(WaveEquationFields.WAVE_SPEED,
                                messages.wave_speed,
                                messages.wave_speed_symbol, 7)
        builder.build_entry_row(WaveEquationFields.LENGTH,
                                messages.length,
                                messages.length_symbol, 8)
        builder.build_entry_row(WaveEquationFields.TIME,
                                messages.time_interval,
                                messages.time_interval_symbol, 9)
        builder.build_entry_row(WaveEquationFields.SAMPLES,
                                messages.samples,
                                messages.samples_symbol, 10)
        self.field_entry_map = builder.get_field_entry_map()

    def build_buttons(self, builder: EquationFormBuilder[WaveEquationFields]):
        self.solve_button = builder.create_button(common_messages.solve, callback=self.solve_equation)
        self.export_button = builder.create_button(common_messages.export, callback=self.export_solution)
        self.render_plot_button = builder.create_button(common_messages.show_plot, callback=self.handle_render_plot)
        self.toggle_animation_button = builder.create_button(common_messages.play,
                                                             callback=self.handle_toggle_animation)
        self.solve_button.grid(row=11, column=2, pady=6, sticky="w")

    def get_equation_metadata(self):
        left_boundary_type = BoundaryType(self.field_entry_map[WaveEquationFields.LEFT_BOUNDARY_TYPE].get())
        right_boundary_type = BoundaryType(self.field_entry_map[WaveEquationFields.RIGHT_BOUNDARY_TYPE].get())
        boundary_conditions = BoundaryConditions(
            BoundaryCondition(left_boundary_type,
                              self.field_entry_map[WaveEquationFields.LEFT_BOUNDARY_VALUES].get()),
            BoundaryCondition(right_boundary_type,
                              self.field_entry_map[WaveEquationFields.RIGHT_BOUNDARY_VALUES].get()))
        source = self.field_entry_map[WaveEquationFields.SOURCE].get()
        return WaveEquationMetadata(
            boundary_conditions,
            float(self.field_entry_map[WaveEquationFields.LENGTH].get()),
            int(self.field_entry_map[WaveEquationFields.SAMPLES].get()),
            float(self.field_entry_map[WaveEquationFields.TIME].get()),
            float(self.field_entry_map[WaveEquationFields.WAVE_SPEED].get()),
            self.field_entry_map[WaveEquationFields.INITIAL_VALUES].get(),
            self.field_entry_map[WaveEquationFields.INITIAL_DERIVATIVES].get(),
            source if source else "0"
        )

    def on_solve(self):
        self.render_plot_button.grid(row=11, column=3, pady=6, sticky="e")
        self.toggle_animation_button.grid(row=12, column=3, pady=6, sticky="e")
        self.export_button.grid(row=13, column=3, pady=6, sticky="e")

    def handle_render_plot(self):
        self.toggle_animation_button.configure(text=common_messages.play)
        self.equation_service.render_current_solution()
        self.canvas.draw()

    def handle_toggle_animation(self):
        self.equation_service.toggle_animation()
        self.canvas.draw()
        if self.equation_service.is_animation_playing():
            self.toggle_animation_button.configure(text=common_messages.pause)
        else:
            self.toggle_animation_button.configure(text=common_messages.play)

    def reset(self):
        self.equation_service.clear_solution()
        self.equation_service.clear_animation()
        self.canvas.draw()
        self.render_plot_button.grid_forget()
        self.toggle_animation_button.grid_forget()
        self.export_button.grid_forget()



nav_bar_font = ("Arial", 10, "bold")
nav_bar_foreground = "#FFFFFF"
nav_bar_background = "#4F5D75"
nav_bar_button_width = 16

details_font = ("Verdana", 10)
details_background = "#BFC0C0"
details_top_margin = 16



from abc import abstractmethod, ABC
from typing import TypeVar, Generic, Optional, Callable

import numpy as np
from matplotlib import cm
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from diffeq_solver_tk.differential_equation_metadata import OrdinaryDifferentialEquationMetadata, BoundedEquationMetadata
from diffeq_solver_tk.util import export_bounded_equation_solution, export_ode_solution

T = TypeVar('T')


class DifferentialEquationService(Generic[T], ABC):
    solve: Callable[[T], np.ndarray]
    main_figure: Optional[Figure]
    metadata: Optional[T] = None
    solution: Optional[np.ndarray] = None

    def __init__(self, solve: Callable[[T], np.ndarray], main_figure: Optional[Figure] = None):
        self.solve = solve
        self.main_figure = main_figure

    def compute_and_update_solution(self, metadata: T):
        self.solution = self.solve(metadata)
        self.metadata = metadata
        self.render_current_solution()

    def clear_solution(self):
        self.metadata = None
        self.solution = None
        self.clear_figure()

    def clear_figure(self):
        self.main_figure.clf()

    @abstractmethod
    def export_solution(self, table_path: str):
        raise NotImplementedError

    @abstractmethod
    def render_current_solution(self):
        raise NotImplementedError


class OrdinaryDifferentialEquationService(DifferentialEquationService[OrdinaryDifferentialEquationMetadata], ABC):
    def render_current_solution(self):
        if self.main_figure is None:
            return
        domain = np.linspace(0, self.metadata.time, self.metadata.samples)
        ax = self.main_figure.add_subplot()
        ax.plot(domain, self.solution)
        ax.set_xlabel("t [time]")
        ax.set_ylabel("x(t)")
        ax.plot()
        ax.set_title("Solution: x(t)")

    def export_solution(self, table_path: str):
        export_ode_solution(self.solution, self.metadata, table_path)


class BoundedEquationService(DifferentialEquationService[BoundedEquationMetadata], ABC):
    __animation: Optional[FuncAnimation] = None
    __is_animation_playing: bool = False

    def render_current_solution(self):
        if self.main_figure is None:
            return
        self.clear_figure()
        self.clear_animation()
        K, N = self.solution.shape
        time_domain = np.linspace(0, self.metadata.time, K)
        space_domain = np.linspace(0, self.metadata.length, N)
        x, t = np.meshgrid(space_domain, time_domain)
        ax = self.main_figure.add_subplot(projection='3d')
        ax.set_xlabel('x [length]')
        ax.set_ylabel('t [time]')
        ax.set_zlabel('u(x, t)')
        surf = ax.plot_surface(x, t, self.solution, cmap=cm.coolwarm, linewidth=0, antialiased=False)
        self.main_figure.colorbar(surf, shrink=0.5, aspect=5)

    def generate_animation(self):
        if self.main_figure is None:
            return
        self.clear_figure()
        K, N = self.solution.shape
        ax = self.main_figure.add_subplot()
        Axes.set_xlim(ax, left=0, right=self.metadata.length)
        Axes.set_ylim(ax, bottom=np.min(self.solution), top=np.max(self.solution))
        ax.set_xlabel("x [length]")
        ax.set_ylabel("u(x, t)")
        ax.set_title("Solution: u(x, t)")
        space_domain = np.linspace(0, self.metadata.length, self.metadata.samples)
        line, = ax.plot(space_domain, self.solution[0], "-o", markersize=4)
        time_text = ax.text(0.82, 0.92, '', transform=ax.transAxes)
        dt = self.metadata.time / K

        def update_plot(k: int):
            line.set_ydata(self.solution[k])
            time_text.set_text("t = %.3f" % (k * dt))
            return line, time_text

        self.__animation = FuncAnimation(self.main_figure, update_plot, frames=K, blit=True, interval=20)

    def toggle_animation(self):
        if not self.__animation:
            self.generate_animation()
            self.__is_animation_playing = True
        elif self.is_animation_playing():
            self.__animation.pause()
            self.__is_animation_playing = False
        else:
            self.__animation.resume()
            self.__is_animation_playing = True

    def is_animation_playing(self):
        return self.__is_animation_playing

    def export_solution(self, table_path: str):
        export_bounded_equation_solution(self.solution, self.metadata, table_path)

    def clear_animation(self):
        if self.main_figure is None or self.__animation is None:
            return
        self.__animation.pause()
        self.__animation = None
        self.__is_animation_playing = False
        self.main_figure.clf()



from abc import abstractmethod
from tkinter import Frame, messagebox
from tkinter.filedialog import asksaveasfilename

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

import diffeq_solver_tk.messages.common_messages as common_messages
import diffeq_solver_tk.tkinter_config as config
from diffeq_solver_tk.differential_equation_service import DifferentialEquationService


class DifferentialEquationForm(Frame):
    canvas: FigureCanvasTkAgg
    equation_service: DifferentialEquationService

    def __init__(self, frame: Frame, canvas: FigureCanvasTkAgg, equation_service: DifferentialEquationService):
        Frame.__init__(self, master=frame)
        self.canvas = canvas
        self.equation_service = equation_service
        self.build_form()
        self.configure(background=config.details_background)

    def solve_equation(self):
        try:
            metadata = self.get_equation_metadata()
            self.equation_service.compute_and_update_solution(metadata)
            self.on_solve()
            self.canvas.draw()
        except Exception as err:
            messagebox.showinfo(common_messages.app_name, err)

    def export_solution(self):
        table_path = asksaveasfilename(filetypes=(("Excel files", "*.xlsx"), ("All files", "*.*")))
        if table_path:
            self.equation_service.export_solution(table_path)

    @abstractmethod
    def get_equation_metadata(self):
        raise NotImplementedError

    @abstractmethod
    def on_solve(self):
        raise NotImplementedError

    @abstractmethod
    def build_form(self):
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        raise NotImplementedError



import argparse
import json
from tkinter import Tk
from typing import List

import diffeq_solver_tk.messages.common_messages as common_messages
from diffeq_solver_tk import __version__
from diffeq_solver_tk.conversion import to_ode_metadata, to_heat_equation_metadata, to_wave_equation_metadata
from diffeq_solver_tk.finite_difference import solve_first_order_ode, solve_second_order_ode, solve_heat_equation, \
    solve_wave_equation
from diffeq_solver_tk.util import export_ode_solution, export_bounded_equation_solution
from .main_view import MainView

first_order_ode_option = "ode1"
second_order_ode_option = "ode2"
heat_equation_option = "heat"
wave_equation_option = "wave"

de_options: List[str] = [
    first_order_ode_option,
    second_order_ode_option,
    heat_equation_option,
    wave_equation_option
]

conversion_map = {
    first_order_ode_option: to_ode_metadata,
    second_order_ode_option: to_ode_metadata,
    heat_equation_option: to_heat_equation_metadata,
    wave_equation_option: to_wave_equation_metadata
}

solve_map = {
    first_order_ode_option: solve_first_order_ode,
    second_order_ode_option: solve_second_order_ode,
    heat_equation_option: solve_heat_equation,
    wave_equation_option: solve_wave_equation
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--version", action="store_true", help="Show version and exit")
    subparsers = parser.add_subparsers(dest="command")
    solve_parser = subparsers.add_parser("solve")
    solve_parser.add_argument("type", type=str, choices=de_options, help="DE type to solve")
    solve_parser.add_argument("infile", type=str, help="DE metadata source file path")
    solve_parser.add_argument("outfile", type=str, help="Solution output file path", default="solution.xlsx")
    args = parser.parse_args()

    if args.version:
        print(f"diffeq-solver-tk {__version__}")
    elif args.command == "solve":
        with open(args.infile) as f:
            query = json.load(f)
            metadata = conversion_map[args.type](query)
            solution = solve_map[args.type](metadata)
            if args.type == first_order_ode_option or args.type == second_order_ode_option:
                export_ode_solution(solution, metadata, args.outfile)
            else:
                export_bounded_equation_solution(solution, metadata, args.outfile)
    else:
        app = Tk(className=common_messages.app_name)
        app.resizable(width=False, height=False)
        # app.iconbitmap('assets/icon.ico')
        app.geometry("1100x600")
        MainView(app).pack(side="top", fill="both", expand=True)
        app.mainloop()


if __name__ == "__main__":
    main()



from typing import Dict, Any

from diffeq_solver_tk.differential_equation_metadata import OrdinaryDifferentialEquationMetadata, HeatEquationMetadata, \
    BoundaryConditions, BoundaryCondition, BoundaryType, WaveEquationMetadata


def to_ode_metadata(query: Dict[str, Any]) -> OrdinaryDifferentialEquationMetadata:
    # TODO: #14
    return OrdinaryDifferentialEquationMetadata(
        query["source"],
        query["initial_derivatives"],
        query["samples"],
        query["time"]
    )


def to_heat_equation_metadata(query: Dict[str, Any]) -> HeatEquationMetadata:
    # TODO: #14
    query_bc = query["boundary_conditions"]
    left_boundary_condition = BoundaryCondition(BoundaryType[query_bc["left"]["type"]], query_bc["left"]["values"])
    right_boundary_condition = BoundaryCondition(BoundaryType[query_bc["right"]["type"]], query_bc["right"]["values"])
    boundary_conditions = BoundaryConditions(left_boundary_condition, right_boundary_condition)
    return HeatEquationMetadata(
        boundary_conditions,
        query["length"],
        query["samples"],
        query["time"],
        query["alpha"],
        query["initial_values"],
        query.get("source", "0")
    )


def to_wave_equation_metadata(query: Dict[str, Any]) -> WaveEquationMetadata:
    # TODO: #14
    query_bc = query["boundary_conditions"]
    left_boundary_condition = BoundaryCondition(BoundaryType[query_bc["left"]["type"]], query_bc["left"]["values"])
    right_boundary_condition = BoundaryCondition(BoundaryType[query_bc["right"]["type"]], query_bc["right"]["values"])
    boundary_conditions = BoundaryConditions(left_boundary_condition, right_boundary_condition)
    return WaveEquationMetadata(
        boundary_conditions,
        query["length"],
        query["samples"],
        query["time"],
        query["wave_speed"],
        query["initial_values"],
        query["initial_derivatives"],
        query.get("source", "0")
    )
